RoIter

Struct RoIter 

pub struct RoIter<'txn, KC, DC, IM = MoveThroughDuplicateValues> { /* private fields */ }
A read-only iterator structure.

Implementations

Source

impl<'txn, KC, DC, IM> RoIter<'txn, KC, DC, IM>

pub fn move_between_keys(self) -> RoIter<'txn, KC, DC, MoveBetweenKeys>

Move on the first value of keys, ignoring duplicate values.

use heed3::types::*;
use heed3::byteorder::BigEndian;

type BEI64 = I64<BigEndian>;

let mut wtxn = env.write_txn()?;
let db = env.database_options()
    .types::<BEI64, BEI64>()
    .flags(DatabaseFlags::DUP_SORT)
    .name("dup-sort")
    .create(&mut wtxn)?;

db.put(&mut wtxn, &68, &120)?;
db.put(&mut wtxn, &68, &121)?;
db.put(&mut wtxn, &68, &122)?;
db.put(&mut wtxn, &68, &123)?;
db.put(&mut wtxn, &35, &120)?;
db.put(&mut wtxn, &0, &120)?;
db.put(&mut wtxn, &42, &120)?;

let mut iter = db.iter(&wtxn)?.move_between_keys();
assert_eq!(iter.next().transpose()?, Some((0, 120)));
assert_eq!(iter.next().transpose()?, Some((35, 120)));
assert_eq!(iter.next().transpose()?, Some((42, 120)));
assert_eq!(iter.next().transpose()?, Some((68, 120)));
assert_eq!(iter.next().transpose()?, None);

drop(iter);
wtxn.commit()?;
pub fn move_through_duplicate_values( self, ) -> RoIter<'txn, KC, DC, MoveThroughDuplicateValues>

Move through key/values entries and output duplicate values.

use heed3::types::*;
use heed3::byteorder::BigEndian;

type BEI64 = I64<BigEndian>;

let mut wtxn = env.write_txn()?;
let db = env.database_options()
    .types::<BEI64, BEI64>()
    .flags(DatabaseFlags::DUP_SORT)
    .name("dup-sort")
    .create(&mut wtxn)?;

db.put(&mut wtxn, &68, &120)?;
db.put(&mut wtxn, &68, &121)?;
db.put(&mut wtxn, &68, &122)?;
db.put(&mut wtxn, &68, &123)?;
db.put(&mut wtxn, &35, &120)?;
db.put(&mut wtxn, &0, &120)?;
db.put(&mut wtxn, &42, &120)?;

let mut iter = db.iter(&wtxn)?.move_through_duplicate_values();
assert_eq!(iter.next().transpose()?, Some((0, 120)));
assert_eq!(iter.next().transpose()?, Some((35, 120)));
assert_eq!(iter.next().transpose()?, Some((42, 120)));
assert_eq!(iter.next().transpose()?, Some((68, 120)));
assert_eq!(iter.next().transpose()?, Some((68, 121)));
assert_eq!(iter.next().transpose()?, Some((68, 122)));
assert_eq!(iter.next().transpose()?, Some((68, 123)));
assert_eq!(iter.next().transpose()?, None);

drop(iter);
wtxn.commit()?;
pub fn remap_types<KC2, DC2>(self) -> RoIter<'txn, KC2, DC2, IM>

Change the codec types of this iterator, specifying the codecs.

pub fn remap_key_type<KC2>(self) -> RoIter<'txn, KC2, DC, IM>

Change the key codec type of this iterator, specifying the new codec.

pub fn remap_data_type<DC2>(self) -> RoIter<'txn, KC, DC2, IM>

Change the data codec type of this iterator, specifying the new codec.

pub fn lazily_decode_data(self) -> RoIter<'txn, KC, LazyDecode<DC>, IM>

Wrap the data bytes into a lazy decoder.